import os

import click

from . import settings, aws, logging_at


@click.group()
@click.option('-c', '--cfg', type=click.File())
@click.option('-l', '--log-level', type=click.Choice(['d', 'i', 'w']), default=os.environ.get('CONFU_LOG') or 'w')
@click.option('-p', '--profile', default=os.environ.get('CONFU_PROFILE') or None)
@click.option('-r', '--region', type=click.Choice(aws.region_names()), default=os.environ.get('CONFU_REGION') or None)
def cli(cfg, log_level, profile, region):
    logging_at(log_level)
    settings.load(
        profile=profile,
        region=region,
        *([cfg] if cfg is not None else [])
    )
    aws.cxn.activate(
        profile_name=settings.profile,
        default_region=settings.region,
    )



"""
Ansible helpers for:

- `dynamic inventories <http://docs.ansible.com/developing_inventory.html>`_
- `lookup plugins <http://docs.ansible.com/developing_plugins.html#lookup-plugins>`_

"""


import collections
import json

import boto.provider
import boto.s3
import click
import yaml

from . import logging_at, aws


__all__ = [
    'AWSLocalInventory',
    'AWSRemoteInventory',
    'S3LookupModule',
    'CitadelLookupModule',
]


class Inventory(object):

    def all(self):
        """
        Returns groups formatted as:

            http://docs.ansible.com/developing_inventory.html

        says.
        """
        raise NotImplementedError

    def one(self, host):
        """
        Returns variables for single host as:

            http://docs.ansible.com/developing_inventory.html

        says.
        """
        raise NotImplementedError

    @property
    def cli(self):
        """
        Creates a:

            http://docs.ansible.com/developing_inventory.html

        compliant command line interface for this inventory, e.g.:

        .. code:: python

            #!/usr/bin/env python

            if __name__ == '__main__':
                inventory.cli()

        See --help for details.
        """
        raise NotImplementedError


class AWSInventory(Inventory):

    def __init__(self, instances, account_alias=None, group_tags=None):
        super(AWSInventory, self).__init__()
        self.instances = instances
        self.account_alias = (
            account_alias or
            # TODO: no aws.cxn.iam().get_account_alias()[0]?
            aws.cxn.iam().get_account_alias()['list_account_aliases_response']['list_account_aliases_result']['account_aliases'][0]
        )
        self.group_tags = []
        for group_tag in group_tags:
            self.group_tag(group_tag)

    def host_for(self, instance):
        raise NotImplementedError

    def instance_for(self, host):
        raise NotImplementedError

    def host_vars(self, instance):
        raise NotImplementedError

    def group_tag(self, tag):
        name, munge, var = None, None, None
        if isinstance(tag, str):
            name = tag
        elif isinstance(tag, (tuple, list)):
            if not 1 <= len(tag) <= 3:
                raise TypeError(
                    'Tag "{0}" must be (name[, munge][, var])'.format(tag)
                )
            name = tag[0]
            if len(tag) > 1:
                munge = tag[1]
                if isinstance(munge, str):
                    spec = munge
                    munge = lambda x: spec.format(value=x)
            if len(tag) > 2:
                var = tag[2]
        else:
            raise TypeError('Tag "{0}" ist not a string or tuple'.format(tag))
        self.group_tags.append((name, munge, var))

    # Inventory

    def all(self):
        groups = collections.defaultdict(
            lambda: dict(hosts=set(), children=set(), vars=dict())
        )

        groups[self.account_alias]['vars']['aws_account'] = self.account_alias

        host_vars = {}
        for instance in self.instances:
            host = self.host_for(instance)
            vars_ = self.host_vars(instance)
            if vars_:
                host_vars[host] = vars_

            # account
            groups[self.account_alias]['children'].add(instance.region.name)
            groups[self.account_alias]['hosts'].add(host)

            # region
            groups[instance.region.name]['hosts'].add(host)
            groups[instance.region.name]['vars']['aws_region'] = (
                instance.region.name
            )

            # region placement
            groups[instance.region.name]['children'].add(instance.placement)
            groups[instance.placement]['hosts'].add(host)
            groups[instance.placement]['vars']['aws_placement'] = instance.placement

            # tagged
            for name, munge, var in self.group_tags:
                if name not in instance.tags:
                    continue
                group_name = value = instance.tags[name]
                if munge:
                    group_name = munge(group_name)
                if isinstance(group_name, str):
                    group_names = [group_name]
                elif isinstance(group_names, list):
                    group_names = group_name
                else:
                    raise TypeError(
                        'Tag "{0}" munge must return string or strings'
                    )
                for group_name in group_names:
                    groups[group_name]['hosts'].add(host)
                    if var:
                        groups[group_name]['vars'][var] = value

        result = {}

        for name, group in list(groups.items()):
            if group['hosts']:
                group['hosts'] = list(group['hosts'])
            else:
                group.pop('hosts', None)
            if group['children']:
                group['children'] = list(group['children'])
            else:
                group.pop('children', None)
            if not group['vars']:
                group.pop('vars', None)
            if group.get('hosts') or group.get('children'):
                result[name] = dict(group)

        result['_meta'] = {
            'hostvars': host_vars
        }

        return result

    def one(self, host):
        instance = self.instance_for(host)
        return self.host_vars(instance) or {}


class AWSLocalInventory(AWSInventory):
    """
    Dynamic inventory for this AWS host, e.g.:

    .. code:: python

        #!/usr/bin/env python

        from confu.ansible import AWSLocalInventory

        inventory = AWSLocalInventory(
            group_tags=[
               'aws:cloudformation:stack-name',
               ('confu:infra-env', 'env-{value}')
               ('confu:app-env', 'env-{value}')
            ],
            group_filters=[
                'julius',
            ]
        )

        if __name__ == '__main__':
            inventory.cli()

    """

    def __init__(self, account_alias=None, group_tags=None):
        super(AWSLocalInventory, self).__init__(
            instances=[aws.this_instance()],
            account_alias=account_alias,
            group_tags=group_tags,
        )

    # AWSInventory

    def host_for(self, instance):
        return '127.0.0.1'

    def instance_for(self, host):
        if '127.0.0.1' == host:
            return self.instances[0]
        raise LookupError('No instances for host "{0}"'.format(host))

    def host_vars(self, instance, default=None):
        return default

    # Inventory

    @property
    def cli(self):

        @click.command()
        @click.option('--list', is_flag=True)
        @click.option('--host', metavar='<hostname>', nargs=1, default=None)
        @click.option('-l', '--log-level', type=click.Choice(['d', 'i', 'w', 'e']), default='e')
        def _cli(list, host, log_level):
            logging_at(log_level)
            if host is not None:
                result = self.one(host)
            else:
                result = self.all()
            print((json.dumps(result, indent=4, sort_keys=True)))

        return _cli


class AWSRemoteInventory(AWSInventory):
    """
    Dynamic inventory for remote AWS host(s), e.g.:

    .. code:: python

        #!/usr/bin/env python

        from confu.ansible import AWSRemoteInventory

        inventory = AWSRemoteInventory(
            'julius',
            instances=(
                aws.instances()
                .regions('us-west-1')
                .filter(('tag:confu:infra', 'julius'))
            ),
            group_tags=[
               'aws:cloudformation:stack-name',
               ('confu:infra-env', 'env-{value}', 'infra_env')
               ('confu:app-env', 'env-{value}', 'app_env')
            ],
        )

        if __name__ == '__main__':
            inventory.cli()

    """

    def __init__(
            self,
            instances,
            account_alias=None,
            group_tags=None,
            vars_metadata_key=None,
            vars_metadata_type=None):
        super(AWSRemoteInventory, self).__init__(
            instances,
            account_alias=account_alias,
            group_tags=group_tags,
        )
        self.vars_metadata_key = vars_metadata_key
        self.vars_metadata_type = vars_metadata_type

    def filter(self, *regions):
        self.instances = self.instances.regions(*regions)

    vars_metadata_types = {
        'application/json': json.loads,
        'application/x-yaml': yaml.load,
    }

    # AWSLocalInventory

    def host_for(self, instance):
        return instance.ip_address or instance.private_ip_address

    def instance_for(self, host):
        instances = list(self.instances.filter(('private_ip_address', host)))
        if not instances:
            instances = list(self.instances.filter(('ip_address', host)))
            if not instances:
                raise LookupError('No instances for host "{0}"'.format(host))
        if len(instances) > 1:
            raise Exception('Multiple instances for host "{0}"'.format(host))
        return instances[0]

    def host_vars(self, instance, default=None):
        if not aws.is_stack_resource(instance):
            return default
        host = self.host_for(instance)
        md = aws.retry()(aws.stack_resource_metadata, instance, None)
        if md is None:
            return default
        vars_ = md
        if self.vars_metadata_key:
            for k in self.vars_metadata_key:
                if k not in vars_:
                    return default
                vars_ = vars_[k]
        if isinstance(vars_, str):
            text_vars = vars_
            if self.vars_metadata_type not in self.vars_metadata_types:
                raise ValueError(
                    'vars_metadata_type="{0}" not supported.'
                    .format(self.vars_metadata_type)
                )
            vars_ = self.vars_metadata_types[self.vars_metadata_type](
                text_vars
            )
            if not isinstance(vars_, dict):
                raise TypeError(
                    'Host "{0}" metadata @ {1}\n{2}\n is not a "{3}" mapping.'
                    .format(
                        host,
                        self.vars_metadata_key,
                        text_vars,
                        self.vars_metadata_type
                    )
                )
        elif not isinstance(vars_, dict):
            raise TypeError(
                'Host "{0}" metadata @ {1} not a mapping.'.format(host)
            )
        if instance.private_ip_address:
            vars_['private_ip_address'] = instance.private_ip_address
        return vars_

    # Inventory

    @property
    def cli(self):

        @click.command()
        @click.option('--list', is_flag=True)
        @click.option('--host', metavar='<hostname>', nargs=1, default=None)
        @click.option('--no-cache', is_flag=True, help='Do not cache.')
        @click.option('--bust', is_flag=True, help='Bust cache first.')
        @click.option('-r', '--regions', type=click.Choice(['us-west-1', 'us-west-2', 'us-east-1']), multiple=True)
        @click.option('-l', '--log-level', type=click.Choice(['d', 'i', 'w', 'e']), default='e')
        def _cli(list, host, no_cache, bust, regions, log_level):
            # TODO: caching w/ no_cache, bust awareness
            logging_at(log_level)
            if regions:
                self.filter(*regions)
            if host is not None:
                result = self.one(host)
            else:
                result = self.all()
            print((json.dumps(result, indent=4, sort_keys=True)))

        return _cli


class S3LookupModule(object):
    """
    Use it like:

    .. code:: python

        import confu

        class LookupModule(confu.ansible.S3LookupModule):

            bucket_var = 'citadel_bucket'

            profile_var = 'citadel_profile'

            region_var = 'citadel_region'

    and then load variable content from s3 in ansible like:

    .. code:: yaml

        newrelic_license_key: "{{ lookup('citadel', '/newrelic/license_key').strip() }}"

    """

    #: Default bucket. Override with "bucket=".
    bucket = None

    #: Variable name with default bucket.
    bucket_var = None

    #: Default connection profile. Override with "profile=".
    profile = None

    #: Variable name with default connection profile.
    profile_var = None

    #: Default connection region. Override with "region=".
    region = None

    #: Variable name with default connection region.
    region_var = None

    #: Object satisfying `collections.MutableMapping` or None for no caching.
    cache = None

    def cache_key(self, bucket_name, key_name):
        return 's3://{0}{1}'.format(bucket_name, key_name)

    def loads(self, content_type, content):
        return content

    def options(self, inject):
        options = {
            'bucket': self.bucket,
            'profile': self.profile,
            'region': self.region,
        }
        if inject:
            for option, var in [
                    ('bucket', self.bucket_var),
                    ('profile', self.profile_var),
                    ('region', self.region_var),
                ]:
                if var and var in inject:
                    options[option] = inject[var]
        return options

    # LookupModule

    def __init__(self, basedir=None, **kwargs):
        self.basedir = basedir

    def run(self, terms, inject=None, **kwargs):
        # XXX: messes with logging
        from . import ansible

        # XXX: https://github.com/ansible/ansible/issues/7370

        terms = ansible.utils.listify_lookup_plugin_terms(terms, self.basedir, inject)

        ret = []

        defaults = self.options(inject)

        for term in terms:
            params = term.split()

            key_name = params[0]

            options = defaults.copy()
            try:
                for param in params[1:]:
                    name, value = param.split('=')
                    assert(name in options)
                    options[name] = value
            except (ValueError, AssertionError) as e:
                raise ansible.errors.AnsibleError(e)

            # cache
            if self.cache is not None:
                cache_key = self.cache_key(options['bucket'], key_name)
                if cache_key in self.cache:
                    contents = self.cache[cache_key]
                    ret.append(contents)
                    continue

            # connect
            profile_name = options['profile']
            cxn = None
            while cxn is None:
                try:
                    if options['region']:
                        cxn = boto.s3.connect_to_region(
                            options['region'], profile_name=profile_name,
                        )
                    else:
                        cxn = boto.connect_s3(profile_name=profile_name)
                except boto.provider.ProfileNotFoundError as e:
                    if profile_name is None:
                        raise ansible.errors.AnsibleError('Unable to connect to s3')
                    profile_name = None
                except (boto.exception.BotoClientError, boto.exception.S3ResponseError) as e:
                    raise ansible.errors.AnsibleError('Unable to connect to s3')

            # read
            # NOTE: lookup deferred (validate=False) so we don't need s3:List* permission
            bucket = cxn.get_bucket(options['bucket'], validate=False)
            key = boto.s3.key.Key(bucket, key_name)
            try:
                contents = key.get_contents_as_string()
            except boto.exception.S3ResponseError as e:
                raise ansible.errors.AnsibleError(
                    'Unable to get s3://{bucket}/{key} contents - {code}, {message}'.format(
                        bucket=options['bucket'],
                        key=key_name,
                        code=e.error_code,
                        message=e.message,
                    )
                )

            # cache
            if self.cache is not None:
                cache_key = self.cache_key(options['bucket'], key_name)
                self.cache[cache_key] = contents

            ret.append(contents)

        return ret



import os

from click._bashcomplete import bashcomplete

from .. import cli


@cli.group()
def shell():
    pass


@shell.command('env')
def env():
    script_path = os.path.join(os.path.dirname(__file__), 'confue.bash')
    print((open(script_path, 'r').read()))


@shell.command('complete')
def complete():
    bashcomplete(cli, 'confu', '_CONFU_COMPLETE', 'source')



import collections
import datetime
import json
import threading
import time

import boto.cloudformation
import boto.iam
import boto.ec2
import boto.s3


def region_names(service='ec2'):
    return [region.name for region in boto.regioninfo.get_regions(service)]


class Connection(threading.local):
    """
    Global per-thread connection map. Typically used via singleton like this:

    .. code: python

        cxn.activate(profile_name='my-profile')

    or, if you don't control the context, like this:

    .. code: python

        with cxn.activate(profile_name='my-profile'):
            pass

    so that subsequent code can just do:

    .. code: python

        s3 = aws.cxn.s3('us-west-1')

    to re-use commonly authenticated connections.
    """

    def __init__(self):
        self.reset()

    def activate(self,
                 access_key=None,
                 secret_key=None,
                 security_token=None,
                 profile_name=None,
                 default_region=None,
        ):
        self.reset()
        self.default_region = default_region
        self.creds = {
            'aws_access_key_id': access_key,
            'aws_secret_access_key': secret_key,
            'security_token': security_token,
            'profile_name': profile_name,
        }

    def reset(self):
        self.default_region = None
        self.creds = {
            'aws_access_key_id': None,
            'aws_secret_access_key': None,
            'security_token': None,
            'profile_name': None,
        }
        self._cloudformation = {}
        self._ec2 = {}
        self._s3 = {}
        self._iam = {}

    def s3(self, region=None):
        return self._connect(region, boto.s3, self._s3)

    def iam(self, region=None):
        return self._connect(region, boto.iam, self._iam)

    def cloudformation(self, region=None):
        return self._connect(region, boto.cloudformation, self._cloudformation)

    def ec2(self, region=None):
        return self._connect(region, boto.ec2, self._ec2)

    def _connect(self, region, service, cache):
        if region is None:
            if self.default_region is None:
                raise ValueError('No default_region yet region=None')
            if self.default_region == 'this':
                identity = boto.utils.get_instance_identity()
                self.default_region = identity['document']['region']
            region = self.default_region
        if region not in cache:
            cxn = service.connect_to_region(region_name=region, **self.creds)
            cache[region] = cxn
        return cache[region]


#: Singleton global per-thread connection map.
cxn = Connection()


class instances(collections.Sequence):
    """
    A sequence of filtered instances, e.g.:

    .. code:: python

        print list(
            instances()
            .regions('us-west-1')
            .filter(('tag:confu:infra', 'fical'))
        )

    """

    def __init__(self, regions=None, filters=None):
        self._regions = regions or []
        self._filters = filters or []
        self._instances = None

    def regions(self, *regions):
        return type(self)(regions, self._filters)

    def filter(self, *filters):
        return type(self)(self._regions, self._filters + list(filters))

    def __getitem__(self, key):
        return self.all()[key]

    def __len__(self):
        return self.len(self.all())

    def all(self):
        if self._instances is None:
            instances = []
            for region_name in self._regions:
                ec2 = cxn.ec2(region_name)
                for instance in ec2.get_only_instances(filters=self._filters):
                    instances.append(instance)
            self._instances = instances
        return self._instances


def this_instance():
    identity = boto.utils.get_instance_identity()
    region_name = identity['document']['region']
    instance_id = identity['document']['instanceId']
    instances = cxn.ec2(region=region_name).get_only_instances(
        instance_ids=[instance_id]
    )
    if not instances:
        raise Exception(
            'No instance with id {0} in region {1}'
            .format(identity['instanceId'], identity['region'])
        )
    return instances[0]


def this_region():
    identity = boto.utils.get_instance_identity()
    region_name = identity['document']['region']
    return region_name


def get_s3_bucket(name, region=None, create=False):
    try:
        return cxn.s3(region).get_bucket(name)
    except boto.exception.S3ResponseError:
        return cxn.s3(region).create_bucket(
            name,
            location={
                'us-west-1': boto.s3.connection.Location.USWest,
                'us-west-2': boto.s3.connection.Location.USWest2,
                'us-east-1': boto.s3.connection.Location.DEFAULT,
                'eu-west-1': boto.s3.connection.Location.EU,
            }[region],
            policy='private',
        )


def url_for_s3_key(key, auth=False):
    if auth:
        return key.generate_url(
            expires_in=datetime.timedelta(days=1).total_seconds()
        )
    return 'https://{host}/{bucket}/{key}'.format(
        host=cxn.s3().server_name(),
        bucket=key.bucket.name,
        key=key.name,
    )


def is_stack_resource(resource):
    return (
        'aws:cloudformation:stack-name' in resource.tags and
        'aws:cloudformation:logical-id' in resource.tags
    )


def describe_stack_resource(resource):
    return cxn.cloudformation().describe_stack_resource(
        resource.tags['aws:cloudformation:stack-name'],
        resource.tags['aws:cloudformation:logical-id']
    )


def stack_resource_metadata(resource, *args):
    result = (
        describe_stack_resource(resource)
        ['DescribeStackResourceResponse']
        ['DescribeStackResourceResult']
        ['StackResourceDetail']
    )
    v = result.get('Metadata', *args)
    if isinstance(v, str):
        v = json.loads(result['Metadata'])
    return v


def retry(sleep=1, max_sleep=60, max_count=10):
    """
    https://github.com/ansible/ansible-modules-core/blob/7fb0605824df897548e5cd3fd00789365c8f20df/cloud/amazon/ec2_elb_lb.py#L404
    """

    def _loop(func, *args, **kwargs):
        count = 0
        while True:
            try:
                return func(*args, **kwargs)
            except boto.exception.BotoServerError as ex:
                if ex.code != 'Throttling' or count == max_count:
                    raise
                cur_sleep = min(max_sleep, sleep * (2 ** count))
                time.sleep(cur_sleep)
                count += 1

    return _loop



import configparser
import json
import logging
import os
import tempfile

import boto.regioninfo
import pilo


logger = logging.getLogger(__name__)


class Region(pilo.fields.String):

    choices = [region.name for region in boto.regioninfo.get_regions('ec2')]

    def __init__(self, *args, **kwargs):
        super(Region, self).__init__(choices=self.choices, *args, **kwargs)


class Default(pilo.Form):

    #: Profile to use.
    profile = pilo.fields.String(default=None)


class AWS(pilo.Form):

    #: Support regions.
    regions = pilo.fields.List(Region(), default=lambda: Region.choices)

    #: Default region.
    default_region = Region(default=lambda: Region.choices[0])

    @default_region.validate
    def default_region(self, value):
        if value in self.regions:
            return True
        msg = '"{0}" must be one of {1}'.format(
            value, ', '.join(['"{0}"'.format(x) for x in self.regions])
        )
        self.ctx.errors.invalid(msg)
        return False


class CFN(pilo.Form):

    #: Format specification for template bucket names. It can reference:
    #: - ``profile``
    #: - ``region`
    bucket_format = pilo.fields.String(default='confu-cfn-{region}')

    #: Template bucket key (i.e. directory).
    bucket_key = pilo.fields.String(default='default')

    #: Format specification for stack names. It can reference:
    #: - any template parameter (e.g. `InfraEnv`).
    #: - ``random` which is just a random string.
    stack_name_format = pilo.fields.String(default='confu-{random}')

    #: Stack tags. Tag values can reference:
    #: - any template parameter (e.g. `InfraEnv`).
    #: Note that these will be prefixed with "confu:".
    stack_tags = pilo.fields.Dict(pilo.fields.String(), pilo.fields.String(), default=dict)

    #: Default parameters.
    parameters = pilo.fields.Dict(pilo.fields.String(), pilo.fields.String(), default=dict)


class PKG(pilo.Form):

    #: Path to source directory. Paths are expanded relative to containing file.
    source_dir = pilo.fields.String(default='./')

    @source_dir.munge
    def source_dir(self, value):
        if not self.ctx.src_path.location:
            return value
        start = os.path.dirname(self.ctx.src_path.location)
        return os.path.abspath(os.path.relpath(value, start))

    #: Format specification for the name of this package. It can reference:
    #: - any `pkg.Source` attribute (e.g. `pkg.Source.dir_name`).
    name = pilo.fields.String(default='{source.dir_name}')

    #: Format specification for the name of this package. It can reference:
    #: - any `pkg.Source` attribute (e.g. `pkg.Source.git_version`).
    version = pilo.fields.String(default='{source.git_version}')

    #: Default file include patterns (mimics GITIGNORE(5)).
    default_includes = pilo.fields.List(pilo.fields.String())

    @default_includes.default
    def default_includes(self):
        from confu import pkg

        return pkg.Package.default_patterns

    #: File include patterns (mimics GITIGNORE(5)).
    includes = pilo.fields.List(pilo.fields.String(), default=list)

    #: Format specification for the directory where package is built. It can
    #: reference:
    #: - any `pkg.Source` attribute (e.g. `pkg.Source.git_version`).
    #: - any `pkg.Package` attribute (e.g. `pkg.Package.name`).
    stage_dir = pilo.fields.String(default=os.path.join(
        tempfile.gettempdir(), 'confu', '{package.name}-{package.version}'
    ))

    #: Format specification for publish package bucket. It can reference:
    #: - ``profile``
    #: - ``region`
    bucket_format = pilo.fields.String(default='confu-pkg')


class Atlas(pilo.Form):

    #: Path to python package defining a collection of troposphere patterns (aka
    #: an atlas). Paths are expanded relative to file containing this setting.
    source_dir = pilo.fields.String(default=None)

    @source_dir.munge
    def source_dir(self, value):
        if not self.ctx.src_path.location:
            return value
        start = os.path.dirname(self.ctx.src_path.location)
        value = os.path.abspath(os.path.relpath(value, start))
        return value


profile = None

region = None

aws = AWS()

cfn = CFN()

pkg = PKG()

atlas = Atlas()


def locations():
    return [path for path in [
        os.path.expanduser('~/.confu.cfg'),
        os.path.join(os.getcwd(), '.confu.cfg'),
    ] if os.path.exists(path)]


def load(profile=None, region=None, globalize=True, *paths):

    def _union(configs, name, profile):
        srcs = []
        name_profile = '{0} {1}'.format(name, profile) if profile else None
        for path, config in configs:
            if name_profile and config.has_section(name_profile):
                srcs.append(pilo.source.ConfigSource(
                    config,
                    section=name_profile,
                    location=path,
                    preserve_whitespace=True,
                ))
            if config.has_section(name):
                srcs.append(pilo.source.ConfigSource(
                    config,
                    section=name,
                    location=path,
                    preserve_whitespace=True,
                ))
        return pilo.source.UnionSource(srcs, merge='combine')

    # load config
    configs = []
    for path in locations() + list(paths):
        if not path:
            continue
        logger.debug('loading "%s"', path)
        config = configparser.ConfigParser()
        config.optionxform = str  # NOTE: preserve case
        config.read(path)
        configs.append((path, config))

    # merge config
    default = Default(_union(configs, 'default', None))
    aws = AWS(_union(configs, 'aws', profile))
    cfn = CFN(_union(configs, 'cfn', profile))
    pkg = PKG(_union(configs, 'pkg', profile))
    atlas = Atlas(_union(configs, 'atlas', profile))

    # derive
    profile = profile or default.profile
    region = region or aws.default_region

    logger.info('settings -\n%s', json.dumps({
        'profile': profile,
        'region': region,
        'aws': aws,
        'cfn': cfn,
        'pkg': pkg,
        'atlas': atlas,
    }, indent=4))
    if globalize:
        # make it so
        globals().update(
            profile=profile,
            region=region,
            aws=aws,
            cfn=cfn,
            pkg=pkg,
            atlas=atlas,
        )
    return dict(
        profile=profile,
        region=region,
        aws=aws,
        cfn=cfn,
        pkg=pkg,
        atlas=atlas,
    )



"""
Helpers for using these infrastructure tools:

- `troposphere <https://github.com/cloudtools/troposphere>`_
- `aws cfn <http://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/Welcome.html>`_
- `ansible <http://docs.ansible.com/>`_

"""
__version__ = '0.2.1'

__all__ = [
    'ansible',
    'aws',
    'cfn',
    'cli',
    'pkg',
    'settings',
]

import imp
import json
import logging
import os
import sys


logger = logging.getLogger(__name__)


def logging_at(level=logging.ERROR):
    if isinstance(level, str):
        levels = {
            'd': logging.DEBUG, 'debug': logging.DEBUG,
            'i': logging.INFO, 'info': logging.INFO,
            'w': logging.WARN, 'warn': logging.WARN, 'warning': logging.WARN,
            'e': logging.ERROR, 'error': logging.ERROR,
        }
        if level not in levels:
            raise ValueError(
                '{0} invalid, expected one of {1}'.format(level, list(levels.keys()))
            )
        level = levels[level]
    elif not isinstance(level, int):
        raise TypeError('level={0!r} must be string or int'.format(level))
    logging.basicConfig(
        level=level,
        format='%(asctime)s : %(levelname)s : %(name)s : %(message)s',
        stream=sys.stderr,
    )


from . import settings
from . import aws
from .cli import cli
from . import ansible
from . import cfn
from . import pkg
from . import shell


@cli.command('cfg')
def cfg():
    summary = {
        'profile': settings.profile,
        'region': settings.region,
        'aws': settings.aws,
        'cfn': settings.cfn,
        'pkg': settings.pkg,
        'atlas': settings.atlas,
    }
    print((json.dumps(summary, indent=4, sort_keys=True)))


class ImportAtlas(object):
    """
    Import hook for ``confu.atlas`` that:

    - loads settings based on environment
    - applies them
    - imports the registered atlas package

    An ``atlas`` is typically just:

    - python package containing common troposphere patterns
    - registered as a package path via ``confu.settings.atlas``

    """

    @classmethod
    def find_module(cls, fullname, path=None):
        if fullname != 'confu.atlas':
            return
        return cls(fullname, path)

    def __init__(self, fullname, path):
        self.fullname = fullname
        self.path = path

    def load_module(self, name):
        settings.load(
            profile=os.environ.get('CONFU_PROFILE') or None,
            region=os.environ.get('CONFU_REGION') or None,
        )
        aws.cxn.activate(
            profile_name=settings.profile,
            default_region=settings.region,
        )
        logger.info('importing "%s" from "%s"', self.fullname, settings.atlas.source_dir)
        module = imp.load_module(
            name, None, settings.atlas.source_dir, ('', 'r', imp.PKG_DIRECTORY)
        )
        return module

sys.meta_path.append(ImportAtlas)



"""
Manages AWS cloudformation templates and stacks as represented by:

- `cfn.Template` and
- `cfn.Stack`

using contextual information in:

- `cfn.Context`

and exposed like:

.. code:: bash

    $ confu cfn --help

"""
from collections import OrderedDict

import errno
import fnmatch
import hashlib
import logging
import json
import os
import random
import string
import subprocess
import sys
import threading
import time
import urllib.request, urllib.error, urllib.parse
import urllib.parse

import boto.s3.key
import click

from . import settings, cli, aws


logger = logging.getLogger(__name__)


class Context(threading.local):

    def __init__(self):
        self._bucket = None

    def reset(self):
        self._bucket = None

    def key_for(self, *names):
        return '/'.join([settings.cfn.bucket_key] + list(names))

    def url_for(self, key, auth=False):
        return aws.url_for_s3_key(key, auth=auth)

    @property
    def bucket(self):
        if self._bucket is None:
            name = settings.cfn.bucket_format.format(**{
                'profile': settings.profile,
                'region': settings.region,
            })
            self._bucket = aws.get_s3_bucket(
                name, region=settings.region, create=True)
        return self._bucket


ctx = Context()


class Template(object):

    @classmethod
    def from_url(cls, url):
        parsed = urllib.parse.urlparse(url, allow_fragments=False)
        if parsed.scheme in ('', 'file'):
            return Template.from_file(parsed.path)
        if parsed.scheme in ('http', 'https'):
            return Template.from_http(url)
        raise ValueError('Unsupported scheme "{0}"'.format(parsed.scheme))

    @classmethod
    def from_file(cls, path):
        if os.access(path, os.X_OK):
            template = TemplateScript(path)
        else:
            template = TemplateRendered(open(path, 'r').read())
        return template

    @classmethod
    def from_http(cls, url):
        resp = urllib.request.urlopen(url)
        code = resp.getcode()
        if code not in (200,):
            raise Exception('GET "{0}" failed - {1}'.format(url, code))
        body = resp.read()
        return TemplateRendered(body)

    @property
    def body(self):
        raise NotImplementedError

    @property
    def name(self):
        return hashlib.sha1(str(self)).hexdigest().decode('utf-8') + '.json'

    @property
    def params(self):
        return self.body.get('Parameters', {})

    @property
    def required(self):
        return dict(
            (name, param)
            for name, param in list(self.params.items())
            if 'Default' not in param
        )

    @property
    def defaults(self):
        return dict(
            (name, param)
            for name, param in list(self.params.items())
            if 'Default' in param
        )

    def remote(self, auth=False):
        key = ctx.bucket.get_key(ctx.key_for(self.name))
        if key is None:
            raise Exception(
                'Not found @ s3://{bucket}/{key}!'
                .format(bucket=ctx.bucket.name, key=ctx.key_for(self.name))
            )
        return ctx.url_for(key, auth=auth)

    @property
    def has_remote(self):
        return ctx.bucket.get_key(ctx.key_for(self.name)) is not None

    def validate(self):
        if not self.has_remote:
            remote = self.upload()
        else:
            remote = self.remote()
        aws.cxn.cloudformation().validate_template(template_url=remote)

    def upload(self, auth=False):
        key = boto.s3.key.Key(ctx.bucket, ctx.key_for(self.name))
        key.set_contents_from_string(
            str(self), headers={'Content-Type': 'application/json'}
        )
        return ctx.url_for(key, auth=auth)

    def __str__(self):
        json_repr = self.body
        # follow AWSCloudFormation/latest/UserGuide/template-anatomy.html
        # to print out the template
        sorted_keys = [
            'AWSTemplateFormatVersion',
            'Description',
            'Parameters',
            'Mappings',
            'Conditions',
            'Resources',
            'Properties',
            'Function',
            'Outputs',
        ]

        def comparator(rhs, lhs):
            rhs_idx = sorted_keys.index(rhs)
            lhs_idx = sorted_keys.index(lhs)
            return -1 if rhs_idx < lhs_idx else lhs_idx

        t = OrderedDict(sorted(list(json_repr.items()),
                               key=lambda x: x[0],
                               cmp=comparator))

        return t.to_dict()

    def stack(self, name, params):
        return Stack.create(name, self, params)


class TemplateScript(Template):

    def __init__(self, script):
        super(TemplateScript, self).__init__()
        self.script = script
        self._body = None

    def render(self):
        if not os.path.isfile(self.script):
            raise Exception('"{0}" does not exist.'.format(self.script))
        if not os.access(self.script, os.X_OK):
            raise Exception('"{0}" is *not* executable.'.format(self.script))
        try:
            raw = subprocess.check_output(self.script, stderr=sys.stderr)
        except OSError as ex:
            if ex.errno == errno.ENOEXEC:
                logger.error(
                    '"%s" failed to execute, '
                    'are you missing "#!/usr/bin/env python"?',
                    self.script
                )
            raise
        try:
            return json.loads(raw)
        except ValueError:
            logger.error(
                '"%s" should print JSON template, got:\n"%s"',
                self.script, raw,
            )
            raise

    @property
    def body(self):
        if self._body is None:
            self._body = self.render()
        return self._body


class TemplateRendered(Template):

    def __init__(self, body):
        try:
            self._body = json.loads(body)
        except ValueError:
            logger.error('Expected JSON template, got:\n%s', body)
            raise

    @property
    def body(self):
        return self._body


class Stack(object):

    @classmethod
    def all(cls, *statuses):
        return list(map(
            cls.from_summary,
            aws.cxn.cloudformation().list_stacks(statuses or None)
        ))

    @classmethod
    def from_resource(cls, resource):
        return cls(name=resource.tags['aws:cloudformation:stack-name'])

    @classmethod
    def from_summary(cls, summary):
        return cls(
            summary.stack_name,
            id=summary.stack_id,
            status=summary.stack_status,
        )

    @classmethod
    def create(cls, name, template, params):
        # expand stack outputs to params
        tmp = []
        for param in params:
            if len(param) == 1:
                for k, v in list(Stack(param[0]).outputs.items()):
                    if k in template.params:
                        tmp.append((k, v))
            else:
                tmp.append(param)
        params = tmp

        # format name
        name = name or settings.cfn.stack_name_format
        name_ctx = {
            'random': ''.join(
                random.choice(string.ascii_lowercase + string.digits)
                for _ in range(6)
            )
        }
        name_ctx.update(
            (name, param['Default'])
            for name, param in list(template.defaults.items())
        )
        name_ctx.update(
            (name, value)
            for name, value in list(settings.cfn.parameters.items())
        )
        name_ctx.update(params)
        logger.debug('"%s".format(%s)', name, name_ctx)
        name = name.format(**name_ctx)

        # validate params
        for key, _ in params:
            if key not in template.params:
                raise ValueError(
                    '"{0}" is not a parameter for {1} '.format(key, name)
                )

        # environment params
        for k, v in list(settings.cfn.parameters.items()):
            if k not in template.params:
                continue
            if any(key == k for key, _ in params):
                continue
            params.append((k, v.format(**{
                'profile': settings.profile,
                'region': settings.region,
            })))

        # create
        url = template.remote()
        tags = {
            'confu:source': 's3://{bucket}/{key}'.format(
                bucket=ctx.bucket.name, key=ctx.key_for(template.name)
            ),
        }
        tag_ctx = dict((name, value) for name, value in params)
        tags.update(
            ('confu:{0}'.format(name), value.format(tag_ctx))
            for name, value in list(settings.cfn.stack_tags.items())
        )
        id = aws.cxn.cloudformation().create_stack(
            stack_name=name,
            template_url=url,
            parameters=params,
            tags=tags,
            capabilities=['CAPABILITY_IAM'],
        )
        return Stack(name, id=id)

    def __init__(self, name, id=None, status=None):
        self.name = name
        self._id = id
        self._status = status
        self._description = None

    def describe(self):
        if self._description is None:
            self._description = aws.cxn.cloudformation().describe_stacks(
                self.name
            )[0]
        return self._description

    def update(self, template, params):
        # expand stack outputs to params
        tmp = []
        for param in params:
            if len(param) == 1:
                for k, v in list(Stack(param[0]).outputs.items()):
                    if k in template.params:
                        tmp.append((k, v))
            else:
                tmp.append(param)
        params = tmp

        # validate params
        for key, _ in params:
            if key not in self.params:
                raise ValueError(
                    '"{0}" is not a parameter for {1} '.format(
                        key, template.name
                    )
                )

        # existing params
        for k, v in list(self.params.items()):
            if k not in template.params:
                continue
            if any(key == k for key, _ in params):
                continue
            params.append((k, v))

        # environment params
        for k, v in list(settings.cfn.parameters.items()):
            if k not in template.params:
                continue
            if any(key == k for key, _ in params):
                continue
            params.append((k, v.format(**{
                'profile': settings.profile,
                'region': settings.region,
            })))

        # update
        url = template.remote()
        tags = {
            'confu:source': 's3://{bucket}/{key}'.format(
                bucket=ctx.bucket.name, key=ctx.key_for(template.name)
            ),
        }
        tag_ctx = dict((name, value) for name, value in params)
        tags.update(
            ('confu:{0}'.format(name), value.format(tag_ctx))
            for name, value in list(settings.cfn.stack_tags.items())
        )
        aws.cxn.cloudformation().update_stack(
            stack_name=self.name, template_url=url, parameters=params,
            tags=tags,
            capabilities=['CAPABILITY_IAM'],
        )

    def delete(self):
        logger.info('deleting stack %s', self.name)
        aws.cxn.cloudformation().delete_stack(self.name)

    @property
    def id(self):
        if self._id:
            return self._id
        return self.describe().stack_id

    @property
    def status(self):
        if self._status:
            return self._status
        return self.describe().stack_status

    @property
    def is_deleted(self):
        return self.status == 'DELETE_COMPLETE'

    def in_progress(self, desired):
        return self.status == '{0}_IN_PROGRESS'.format(
            desired.partition('_')[0]
        )

    @property
    def params(self):
        return dict(
            (param.key, param.value)
            for param in self.describe().parameters
        )

    @property
    def tags(self):
        return self.describe().tags

    @property
    def outputs(self):
        return dict(
            (output.key, output.value) for output in self.describe().outputs
        )

    def reset(self):
        self._description = None
        self._id = None
        self._status = None

    @property
    def template(self):
        body = aws.cxn.cloudformation().get_template(self.name)
        return TemplateRendered(body)


# cli.cfn

@cli.group('cfn')
def cfn():
    pass


@cfn.command('validate')
@click.argument('template_url')
def validate(template_url):
    template = Template.from_url(template_url)
    template.validate()


@cfn.command('render')
@click.argument('template_url')
def render(template_url):
    template = Template.from_url(template_url)
    print((str(template)))


@cfn.command('params')
@click.argument('template_url')
def parameters(template_url):
    template = Template.from_url(template_url)
    print((json.dumps(template.params, indent=4, sort_keys=True)))


@cfn.command('defaults')
@click.argument('template_url')
def defaults(template_url):
    template = Template.from_url(template_url)
    print((json.dumps(template.defaults, indent=4, sort_keys=True)))


@cfn.command('required')
@click.argument('template_url')
def required(template_url):
    template = Template.from_url(template_url)
    print((json.dumps(template.required, indent=4, sort_keys=True)))


@cfn.command('upload')
@click.argument('template_url')
@click.option('--auth/--no-auth', default=False)
def upload(template_url, auth):
    template = Template.from_url(template_url)
    print((template.upload(auth=auth)))


@cfn.command('uploaded')
@click.argument('template_url')
@click.option('--auth/--no-auth', default=False)
def uploaded(template_url, auth):
    print((Template.from_url(template_url).remote(auth=auth)))


@cfn.command('create')
@click.argument('template_url')
@click.argument('param', nargs=-1)
@click.option('-n', '--stack-name')
def create(template_url, param, stack_name):
    params = [tuple(param.split('=')) for param in param]
    template = Template.from_url(template_url)
    if not template.has_remote:
        template.upload()
    print((template.stack(name=stack_name, params=params).name))


@cfn.command('show')
@click.argument('stack_name')
@click.option('--deleted/--no-deleted', default=False)
def show(stack_name, deleted):

    def _summarize(stack):
        summary = {
            'name': stack.name,
            'id': stack.id,
            'status': stack.status,
        }
        if not stack.is_deleted:
            summary.update(
                params=stack.params,
                outputs=stack.outputs,
                tags=stack.tags,
            )
        return summary

    if any(c in stack_name for c in '%?*'):
        summary = []
        for stack in Stack.all():
            if not deleted and stack.is_deleted:
                continue
            if not fnmatch.fnmatch(stack.name, stack_name):
                continue
            summary.append(_summarize(stack))
    else:
        stack = Stack(stack_name)
        summary = _summarize(stack)
    print((json.dumps(summary, indent=4, sort_keys=True)))


@cfn.command('wait-for')
@click.argument('stack_name')
@click.argument('status', type=click.Choice(['created', 'deleted', 'updated']))
@click.option('--poll', type=int, default=1)
@click.option('-t', '--timeout', type=int, default=None)
@click.pass_context
def wait_for(ctx, stack_name, status, poll, timeout):
    stack = Stack(stack_name)
    desired = {
        'created': 'CREATE_COMPLETE',
        'deleted': 'DELETE_COMPLETE',
        'updated': 'UPDATE_COMPLETE',
    }[status]
    if timeout:
        timeout = time.time() + timeout
    while True:
        if stack.status == desired:
            break
        if not stack.in_progress(desired):
            raise Exception(
                'Bailing on {0} with status {1}'.format(
                    stack.name, stack.status
                )
            )
        if timeout and time.time() > timeout:
            raise Exception(
                'Timeout on {0} with status {1}'.format(
                    stack.name, stack.status
                )
            )
        time.sleep(poll)
        stack = Stack(stack_name)
    ctx.invoke(show, stack.name)


@cfn.command('delete')
@click.argument('stack_names', nargs=-1)
def delete(stack_names):
    for stack_name in stack_names:
        if any(c in stack_name for c in '%?*'):
            for stack in Stack.all():
                if stack.is_deleted:
                    continue
                if not fnmatch.fnmatch(stack.name, stack_name):
                    continue
                stack.delete()
        else:
            Stack(stack_name).delete()


@cfn.command('update')
@click.argument('stack_name')
@click.argument('template_url')
@click.argument('param', nargs=-1)
def update(stack_name, template_url, param):
    params = [tuple(param.split('=')) for param in param]
    template = Template.from_url(template_url)
    if not template.has_remote:
        template.upload()
    Stack(stack_name).update(template=template, params=params)



import re
import setuptools


extras_require = {
    'tests': [
        'mock >=1.0,<2.0',
        'pytest >=2.5.2,<3',
        'pytest-cov >=1.7,<2',
    ]
}

setuptools.setup(
    name='py3-con-fu',
    version=(
        re
        .compile(r".*__version__ = '(.*?)'", re.S)
        .match(open('confu/__init__.py').read())
        .group(1)
    ),
    url='https://github.com/bninja/confu',
    license='BSD',
    author='Egon Spengler',
    author_email='egon@gb.com',
    description='Fu for con.',
    long_description=open('README.rst').read(),
    packages=setuptools.find_packages('.', exclude=('tests', 'tests.*')),
    platforms='any',
    install_requires=[
        'ansible >= 2,<3',
        'boto >= 2.4.6,<3',
        'troposphere >= 2.6,<2.7',
        'click',
        'pilo >=0.4,<0.5',
        'virtualenv',
        'virtualenv-relocate',
    ],
    extras_require=extras_require,
    tests_require=extras_require['tests'],
    scripts=['bin/confu'],
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ]
)



"""
Builds Ansible *site* packages that can be used to configure a machine. They
consist of:

- `ansible playbooks <http://docs.ansible.com/playbooks_best_practices.html#content-organization>`_

and the:

- relocatable python `virtual environment <https://virtualenv.pypa.io/en/latest/>_`

needed to run them. Packages are described by:

    - `Package`
    - `Source`
    - `Pattern`

the build process by:

    - `Makefile`

and exposed to you like:

.. code:: bash

    $ confu pkg --help

You can use built archives directly to configure a machine, e.g. for
``my-conf-0.1.2.tar.gz``:

.. code:: bash

    local$ scp my-conf-0.1.2.tar.gz me@host:
    local$ tar xvf me@host
    local$ ssh me@host
    host$ tar xvf my-conf-0.1.2.tar.gz -C /opt/
    host$ cd /opt/my-conf-0.1.2/
    host$ ./env/bin/ansible-playbook site.yml -i inventories/pull -c local

or use them to build distribution packages (e.g. RPMs, DEBs).

.. note::

    Archives are tied to your build environment (e.g. distro version,
    architecture, etc) so target machines to configure need to be similar.

"""
import distutils
import errno
import fnmatch
import json
import logging
import os
import re
import string
import subprocess
import sys
import tempfile

import boto.s3
import click

from . import settings, aws, cli


logger = logging.getLogger(__name__)


class Package(object):

    default_patterns = [
        'group_vars/',
        'host_vars/',
        'roles/',
        '/ansible.cfg',
        '!*/ansible.cfg',
        '*.yml',
        '!.project',
        '!*.git',
        '!*.pyc',
        '!*.pyo',
        '!*.git*',
        '!*.travis.yml',
        '!*.md',
        '!Vagrantfile',
        '!*/test/',
        '!test.yml',
    ]

    def __init__(self,
                 source_dir,
                 name='{source.dir_name}',
                 version='{source.git_version}',
                 stage_dir=None,
                 patterns=None,
                 bucket_format=None,
        ):
        self.source = Source(os.path.abspath(source_dir))
        self.name = name.format(source=self.source)
        self.version = version.format(source=self.source)
        self.stage_dir = (
            stage_dir or
            os.path.join(
                tempfile.gettempdir(),
                'confu',
                '{package.name}-{package.version}'
            )
        ).format(source=self.source, package=self)
        self.patterns = [
            Pattern.from_expression(expression)
            for expression in patterns or self.default_patterns
        ]
        self.bucket_format = bucket_format

    def include(self, path, is_file):
        include = None
        for pattern in self.patterns:
            if pattern.match(path, is_file):
                include = not pattern.negate
                if pattern.negate:
                    break
        return include

    @property
    def manifest(self):
        paths = []
        for root, dirs, files in os.walk(self.source.dir):
            # match files
            for file in files:
                path = os.path.relpath(os.path.join(root, file), self.source.dir)
                if self.include(path, is_file=True) is True:
                    paths.append(path)

            # recurse prune
            recurse = []
            for dir in dirs:
                path = os.path.relpath(os.path.join(root, dir), self.source.dir)
                if self.include(path, is_file=False) is not False:
                    recurse.append(dir)
            dirs[:] = recurse

        return paths

    @property
    def makefile(self):
        return Makefile(self)

    def install_deps(self):
        """Install dependency roles

        """
        if not os.path.exists('ansible-requirements.yml'):
            logger.info('ansible-requirements.yml does not exist')
            return
        logger.info('installing dependency roles')
        subprocess.check_call([
            'ansible-galaxy', 'install',
            '-r', 'ansible-requirements.yml',
            '-p', '.',
            '--force',
        ])

    def stage(self):
        if not os.path.exists(self.stage_dir):
            logger.info('creating  stage dir %s', self.stage_dir)
            os.makedirs(self.stage_dir)
        manifest_file = os.path.join(self.stage_dir, '.manifest')
        with open(manifest_file, 'w+') as fo:
            logger.info('writing file manifest to %s', manifest_file)
            fo.writelines([line + '\n' for line in self.manifest])
        make_file = os.path.join(self.stage_dir, 'Makefile')
        with open(make_file, 'w+') as fo:
            logger.info('writing Makefile to %s', make_file)
            fo.write(self.makefile.render())

    @property
    def distribution(self):
        return os.path.join(
            self.source.dir,
            '{name}-{version}.tar.gz'.format(name=self.name, version=self.version)
        )

    @property
    def bucket_name(self):
        return self.bucket_format.format(**{
            'profile': settings.profile,
            'region': settings.region,
        })

    @property
    def archive_bucket(self):
        return aws.get_s3_bucket(
            self.bucket_name, region=settings.region, create=True,
        )

    def archive(self, always=False):
        if not os.path.isfile(self.distribution):
            if not always:
                raise OSError((errno.ENOENT, self.distribution))
            logger.info('"%s" not found, trying to create', self.distribution)
            self.stage()
            self.makefile.run('all')
            return self.archive(always=False)
        key_name = os.path.basename(self.distribution)
        bucket = self.archive_bucket
        key = boto.s3.key.Key(bucket, key_name)
        with open(self.distribution, 'rb') as fo:
            logger.info(
                'uploading "%s" to s3://%s/%s',
                self.distribution, self.archive_bucket.name, key_name
            )
            key.set_contents_from_file(fo)
        return aws.url_for_s3_key(key, auth=True)

    @property
    def archived_keys(self):
        prefix = '{0}-'.format(self.name)
        return [
           key for key in self.archive_bucket.list()
           if key.name.startswith(prefix)
        ]

    def archived_key(self, name):
        return self.archive_bucket.get_key(name)

    def archived_url(self, key, auth=False):
        if isinstance(key, str):
            key = self.archived_key(key)
        return aws.url_for_s3_key(key, auth)


class Pattern(object):

    @classmethod
    def from_expression(self, value):
        negate = False
        if value.startswith('!'):
            value = value[1:]
            negate = True
        if value.endswith('/'):
            value = value[:-1]
            cls = PrefixPattern
        else:
            cls = GlobPattern
        return cls(value, negate=negate)

    def __init__(self, negate):
        self.negate = negate

    def match(self, path, is_file):
        raise NotImplementedError

    @property
    def expression(self):
        pass

    def __unicode__(self):
        return self.expression

    def __str__(self):
        return self.__unicode__().encode('utf-8')


class GlobPattern(Pattern):

    def __init__(self, pattern, negate):
        self.pattern = pattern
        self.re = re.compile(fnmatch.translate(pattern.lstrip('/')))
        self.name_only = '/' not in pattern
        super(GlobPattern, self).__init__(negate)

    def match(self, path, is_file):
        if self.name_only:
            path = os.path.basename(path)
        return self.re.match(path) is not None

    @property
    def expression(self):
        return ('!' if self.negate else '') + self.pattern


class PrefixPattern(Pattern):

    def __init__(self, prefix, negate):
        self.prefix = prefix
        super(PrefixPattern, self).__init__(negate)

    def match(self, path, is_file):
        if is_file:
            return path.startswith(self.prefix)
        return path.startswith(self.prefix) or path + '/' == self.prefix

    @property
    def expression(self):
        return ('!' if self.negate else '') + self.prefix + '/'


class Source(object):

    def __init__(self, dir):
        self.dir = dir

    @property
    def dir_name(self):
        return os.path.basename(self.dir)

    @property
    def git_version(self):
        cmd = ['git', 'describe', '--always']
        logger.info('git_version - "%s"', ' '.join(cmd))
        return subprocess.check_output(cmd, cwd=self.dir).strip()


class Makefile(object):

    template = string.Template("""\
NAME = $name
VER = $version

SOURCEDIR = $source_dir

VENV = env
VENVBIN = $virtualenv_bin
VENVREQS = $$(SOURCEDIR)/requirements.txt
VENVRELOCATEBIN = $virtualenv_relocate_bin

STAGEDIR = $stage_dir
MANIFEST = $$(STAGEDIR)/.manifest
BUILDDIR = $$(STAGEDIR)/build

DISTDIR = $$(STAGEDIR)/dist
DISTNAME = $$(NAME)-$$(VER)
DISTFILE = $$(DISTNAME).tar.gz

.PHONY: build-files build-venv dist-sync dist-clean all clean

$$(BUILDDIR):
    mkdir -p $$@

build-files: $$(BUILDDIR)
    rsync -av --files-from=$$(MANIFEST) . $$(BUILDDIR)

build-venv: $$(BUILDDIR)
    $$(VENVBIN) $$(BUILDDIR)/$$(VENV)
    $$(BUILDDIR)/$$(VENV)/bin/pip install -r $$(VENVREQS)
    $$(BUILDDIR)/$$(VENV)/bin/pip uninstall pip setuptools -y
    $$(VENVBIN) --relocatable $$(BUILDDIR)/$$(VENV)
    $$(VENVRELOCATEBIN) $$(BUILDDIR)/$$(VENV) -ld

$$(DISTDIR):
    mkdir -p $$@

dist-sync: build-files build-venv $$(DISTDIR)
    rsync -av --exclude='*pyc' --exclude='*pyo' $$(BUILDDIR)/ $$(DISTDIR)/$$(DISTNAME)

$$(DISTFILE): dist-sync
    tar -czf $$(DISTDIR)/$$(DISTFILE) -C $$(DISTDIR) $$(DISTNAME)

dist-clean:
    rm -rf $$(DISTDIR)
    rm -rf $$(SOURCEDIR)/$$(DISTFILE)

$$(SOURCEDIR)/$$(DISTFILE): $$(DISTFILE)
    cp $$(DISTDIR)/$$(DISTFILE) $$(SOURCEDIR)/$$(DISTFILE)

all: $$(SOURCEDIR)/$$(DISTFILE)

clean: dist-clean
    rm -rf $$(STAGEDIR)

""".replace(' ' * 4, '\t'))

    def __init__(self, package):
        self.package = package

    def render(self):

        def find_executable(name):
            # if invoked from an interpreter w/o fiddling w/ path,
            # try to search distribution first, then fallback to default
            for search_path in [os.path.dirname(sys.executable), None]:
                executable = distutils.spawn.find_executable(name,
                                                             path=search_path)
                if executable:
                    return executable
            raise IOError(errno.ENOENT, 'No such file or directory', name)

        return self.template.substitute(
            name=self.package.name,
            version=self.package.version,
            source_dir=self.package.source.dir,
            stage_dir=self.package.stage_dir,
            virtualenv_bin=find_executable('virtualenv'),
            virtualenv_relocate_bin=find_executable('virtualenv-relocate')
        )

    def __unicode__(self):
        return self.render()

    def __str__(self):
        return self.__unicode__().encode('utf-8')

    def run(self, target):
        cmd = [
            'make',
            '-f', os.path.join(self.package.stage_dir, 'Makefile'),
            target
        ]
        logger.info('run(target="%s") - "%s"', target, ' '.join(cmd))
        return subprocess.call(cmd, stdout=sys.stdout, stderr=sys.stderr)


# cli.pkg

@cli.group('pkg')
@click.option('-n', '--name')
@click.option('-v', '--version')
@click.option('-s', '--source-dir')
@click.option('-b', '--bucket')
@click.pass_context
def pkg(ctx, name, version, source_dir, bucket):
    ctx.package = Package(
        source_dir=source_dir or settings.pkg.source_dir,
        name=name or settings.pkg.name,
        version=version or settings.pkg.version,
        patterns=settings.pkg.includes + settings.pkg.default_includes,
        bucket_format=bucket or settings.pkg.bucket_format
    )


@pkg.command('info')
@click.pass_context
def info(ctx):
    package = ctx.parent.package
    info = {
        'name': package.name,
        'version': package.version,
        'source': {
            'dir': package.source.dir,
        },
        'patterns': list(map(str, package.patterns)),
        'stage_dir': package.stage_dir,
    }
    print((json.dumps(info, indent=4, sort_keys=True)))


@pkg.command('makefile')
@click.pass_context
def makefile(ctx):
    print((ctx.parent.package.makefile))


@pkg.command('manifest')
@click.pass_context
def manifest(ctx):
    for path in ctx.parent.package.manifest:
        print(path)


@pkg.command('init')
@click.pass_context
def init(ctx):
    ctx.parent.package.install_deps()


@pkg.command('stage')
@click.pass_context
def stage(ctx):
    ctx.parent.package.stage()


@pkg.command('build')
@click.pass_context
def build(ctx):
    ctx.parent.package.stage()
    sys.exit(ctx.parent.package.makefile.run('all'))


@pkg.command('clean')
@click.pass_context
def clean(ctx):
    ctx.parent.package.stage()
    sys.exit(ctx.parent.package.makefile.run('clean'))


@pkg.command('archive')
@click.option('-a', '--always', is_flag=True, default=False)
@click.pass_context
def archive(ctx, always):
    print((ctx.parent.package.archive(always=always)))


@pkg.command('archived')
@click.argument('name', required=False)
@click.option('--auth/--no-auth', default=None)
@click.pass_context
def archived(ctx, name, auth):

    def _url(key):
        print((package.archived_url(key.name, auth)))

    def _name(key):
        print((key.name))

    render = {
        True: _url,
        False: _url,
        None: _name,
    }[auth]

    package = ctx.parent.package

    if name is None:
        for key in package.archived_keys:
            render(key)
    elif any(c in name for c in '%?*'):
        for key in package.archived_keys:
            if fnmatch.fnmatch(key.name, name):
                render(key)
    else:
        key = package.archived_key(name)
        render(key, auth=auth)
